/* Stolen from ekg2 code,
 * (c) by by Wojtek Kaniewski, Dawid Jarosz */

#include "dynstuff.h"

void *list_add_sorted(list_t *list, void *data, int alloc_size,
		int (*comparision)(void *, void *), int dont_allow_dups)
{
	list_t new, tmp;

	if (!list) {
		errno = EFAULT;
		return NULL;
	}

	new = xmalloc(sizeof(struct list));

	new->data = data;
	new->next = NULL;
	new->prev = NULL;

	if (alloc_size) {
		new->data = xmalloc(alloc_size);
		memcpy(new->data, data, alloc_size);
	}

	if (!(tmp = *list)) {
		*list = new;
	} else {
		if (!comparision) {
			while (tmp->next)
				tmp = tmp->next;
			tmp->next = new;
		} else {
			list_t prev = NULL;

			while (comparision(new->data, tmp->data) > 0) {
				prev = tmp;
				tmp = tmp->next;
				if (!tmp)
					break;
			}
			if (dont_allow_dups)
				if (tmp && !comparision(new->data, tmp->data))
				{
					if (alloc_size)
						xfree(new->data);
					xfree(new);
					return NULL;
					/*
					return tmp->data;
					*/
				}

			if (!prev) {
				tmp = *list;
				*list = new;
				new->next = tmp;
			} else {
				prev->next = new;
				new->next = tmp;
			}
		}
	}

	return new->data;
}

void *list_add(list_t *list, void *data, int alloc_size)
{
	return list_add_sorted(list, data, alloc_size, NULL, 0);
}

int list_remove(list_t *list, void *data, int free_data)
{
	list_t tmp, last = NULL;

	if (!list) {
		errno = EFAULT;
		return -1;
	}

	tmp = *list;
	if (tmp->data == data) {
		*list = tmp->next;
	} else {
		for (; tmp && tmp->data != data; tmp = tmp->next)
			last = tmp;
		if (!tmp) {
			errno = ENOENT;
			return -1;
		}
		last->next = tmp->next;
	}

	if (free_data)
		xfree(tmp->data);
	xfree(tmp);

	return 0;
}

int list_count(list_t list)
{
	int count = 0;

	for (; list; list = list->next)
		count++;

	return count;
}

int list_destroy(list_t list, int free_data)
{
	list_t tmp;

	while (list) {
		if (free_data)
			xfree(list->data);

		tmp = list->next;

		xfree(list);

		list = tmp;
	}

	return 0;
}

#define fix(s) ((s) ? (s) : "")

/* OUT OF MEM HANDLER */
void oom_handler(int s)
{
	fprintf(stderr, "\n\n"
			"*** Brak pamici ***\n"
			"Konczymy natychmiastowo\n"
			"[dodatkowe: %d\n]"
			"\n\n", s);
	exit(1);
}

void *xcalloc(size_t nmemb, size_t size)
{
	void *tmp = calloc(nmemb, size);

	if (!tmp)
		oom_handler(size);

	return tmp;
}

void *xmalloc(size_t size)
{
	void *tmp = malloc(size);

	if (!tmp)
		oom_handler(size);

	memset(tmp, 0, size);

	return tmp;
}

void xfree(void *ptr)
{
	if (ptr)
		;
	free(ptr);
}

void xfreen(void *ptr)
{
	free(*(void **)ptr);
	*(void **)ptr = NULL;
}

void *xrealloc(void *ptr, size_t size)
{
	void *tmp = realloc(ptr, size);

	if (!tmp)
		oom_handler(size);

	return tmp;
}

char *xstrdup(const char *s)
{
	char *tmp;

	if (!s)
		return NULL;

	if (!(tmp = strdup(s)))
		oom_handler(0);

	return tmp;
}

size_t xstrlen(const char *s)
{
	return strlen(fix(s));
}

char *xstrcpy(char *dest, const char *src)
{
	return strcpy(dest, fix(src));
}

char *xstrncpy(char *dest, const char *src, size_t n)
{
	return strncpy(dest, fix(src), n);
}

char *xstrchr(char *s, char c)
{
	return strchr(fix(s), c);
}

